# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import aiohttp
import textwrap
from dataclasses import dataclass
from fediverse_pasture.types import ApplicationAdapterForActor


@dataclass
class VerifyActorRunner:
    """Class to run query the verify runner application for
    various applications"""

    application: ApplicationAdapterForActor
    verify_actor_url: str = "http://pasture_verify_actor/"
    session: aiohttp.ClientSession | None = None

    async def run(self) -> list:
        """Runs the query. Returns a list containing the results
        for each application"""
        if not self.session:
            async with aiohttp.ClientSession() as session:
                return await self.run_for_session(session)

        return await self.run_for_session(self.session)

    async def run_for_session(self, session) -> dict:
        async with session.post(
            self.verify_actor_url,
            data=aiohttp.FormData({"actor_uri": self.application.actor_uri}),
            headers={
                "content_type": "application/x-www-form-urlencoded",
                "accept": "application/json",
            },
        ) as response:
            data = await response.json()
            return {
                "application_name": self.application.application_name,
                "verify_actor_table": data["result"],
                "messages": data["messages"],
            }


def x_or_not(value):
    if value:
        return "X"
    return " "


def format_verify_actor_result(result, prefix="###"):
    """Routine to convert the obtained result into a markdown string"""
    lines = []
    lines.append(f"{prefix} Verify Actor")
    lines.append("")
    for entry in result:
        lines.append(f'''=== "{entry["application_name"]}"''')
        lines.append("")
        table = [
            ("| Name   | GET Actor | POST Inbox |"),
            ("| ------ | --------- | ---------- |"),
        ]
        for name, value in entry["verify_actor_table"].items():
            if name != "webfinger":
                x_get = x_or_not(value["get_actor"])
                x_post = x_or_not(value["post_inbox"])
                table.append(f"""| {name} | {x_get} | {x_post} |""")

        lines.append(textwrap.indent("\n".join(table), " " * 4))
        lines.append("")

    lines.append("")
    lines.append(f"{prefix} Messages")
    lines.append("")
    for entry in result:
        lines.append(f'''=== "{entry["application_name"]}"''')
        lines.append("")
        lines.append("    ```json")
        lines.append(textwrap.indent(entry["messages"], " " * 4))
        lines.append("    ```")
        lines.append("")

    return "\n".join(lines)



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import pytest
from quart import Quart

from .image_provider_blueprint import image_provider_blueprint


@pytest.fixture
async def image_test_client():
    app = Quart(__name__)
    app.register_blueprint(image_provider_blueprint)

    yield app.test_client()


async def test_can_get_image_at_url(image_test_client):
    response = await image_test_client.get("/aaa111")

    assert response.status_code == 200



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import json
import tomllib
import tomli_w
from contextlib import asynccontextmanager

from tortoise import Tortoise

from .models import TestRecord
from .entry import Entry


class ResultStore:
    async def add_result(self, test_name: str, application_name: str, data: dict):
        """Adds a result to the database. The pairs (test_name, application_name)
        are assumed to be unique. If data already exists, it is overwritten with
        the new data.

        :param test_name: Name of the test
        :param application_name: Name of the application
        :param data: Data to add
        """
        await TestRecord.update_or_create(
            test_name=test_name,
            application_name=application_name,
            defaults={"data": data},
        )

    async def delete_record(self, test_name: str, application_name: str):
        """Deletes database record if exists

        :param test_name: Name of the test
        :param application_name: Application to delete the result for
        """
        record = await TestRecord.get_or_none(
            test_name=test_name,
            application_name=application_name,
        )
        if record:
            await record.delete()

    async def results_for_test(self, test_name: str) -> list:
        """Retrieves the results for a given test_name"""
        result = await TestRecord.filter(test_name=test_name).all()

        return [{"application_name": x.application_name, **x.data} for x in result]

    async def entry_for_test(self, test_name: str) -> Entry:
        """Returns an Entry for test_name. We note here that an
        Entry contains the results for each application the test result
        is available for.

        :param test_name: name of the test entry
        :return: Entry from record
        """
        result = await TestRecord.filter(test_name=test_name).all()

        return Entry({x.application_name: x.data for x in result})

    async def as_list(self) -> list:
        records = await TestRecord.filter().order_by("test_name", "application_name")

        return [
            {
                "test_name": record.test_name,
                "application_name": record.application_name,
                "data": json.dumps(record.data, indent=2),
            }
            for record in records
        ]

    async def load(self, filename: str = "test_results.toml"):
        """Deletes the current content from the database, then loads
        the content from filename to it.

        :param filename: filename to load from"""

        await TestRecord.filter().delete()

        with open(filename, "rb") as fp:
            data = tomllib.load(fp)

        for x in data["entries"]:
            await self.add_result(
                test_name=x["test_name"],
                application_name=x["application_name"],
                data=json.loads(x["data"]),
            )

    async def save(self, filename: str = "test_results.toml"):
        """Saves the content of the database to the file given by filename

        :param filename: filename to save to
        """

        with open(filename, "wb") as fp:
            tomli_w.dump({"entries": await self.as_list()}, fp, multiline_strings=True)


@asynccontextmanager
async def with_store(db_url: str = "sqlite://test_results.sqlite") -> ResultStore:
    """Initializes the database and returns a ResultStore. Usage:

    ```python
    async with with_store() as store:
        await store.add_result(...)
        ...
        await store.results_for_test(...)
    ```

    :param db_url: Database url string
    """
    await Tortoise.init(
        config={
            "connections": {"default": db_url},
            "apps": {
                "models": {
                    "models": [
                        "fediverse_pasture.runner.models",
                    ],
                    "default_connection": "default",
                },
            },
        },
    )
    await Tortoise.generate_schemas()

    yield ResultStore()

    await Tortoise.close_connections()



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import pytest
import json
from datetime import datetime
from unittest.mock import AsyncMock

from fediverse_pasture.data_provider import DataProvider, bovine_actor_for_actor_data

from . import ActivitySender


@pytest.fixture
def activity_sender():
    dp = DataProvider.generate(with_possible_actors=False)

    bovine_actor, actor_object = bovine_actor_for_actor_data(
        "http://localhost/actor", dp.one_actor
    )

    yield ActivitySender.for_actor(bovine_actor, actor_object)


def test_activity_sender(activity_sender):
    activity_sender.init_create_note(lambda x: {**x, "content": "text"})

    assert isinstance(activity_sender.published, datetime)
    assert isinstance(activity_sender.note, dict)

    obj = activity_sender.note

    assert obj.get("content") == "text"

    text = json.dumps(activity_sender.note)

    assert len(text) > 20


async def test_activity_sender_modifying_to(activity_sender):
    activity_sender.init_create_note(lambda x: {**x})
    assert activity_sender.note.get("to") == [
        "https://www.w3.org/ns/activitystreams#Public"
    ]

    activity_sender.init_create_note(lambda x: {**x, "to": ["Public"]})
    assert activity_sender.note.get("to") == ["Public"]

    activity_sender.bovine_actor.post = AsyncMock()
    activity_sender.bovine_actor.get = AsyncMock(
        return_value={"inbox": "https://remote.example/inbox"}
    )

    await activity_sender.send("https://remote.example")

    activity_sender.bovine_actor.get.assert_awaited_once()
    assert activity_sender.bovine_actor.get.await_args == [("https://remote.example",)]
    activity_sender.bovine_actor.post.assert_awaited_once()

    args = activity_sender.bovine_actor.post.await_args[0]

    assert args[0] == "https://remote.example/inbox"
    assert set(args[1]["to"]) == {"https://remote.example", "Public"}


async def test_activity_sender_mention(activity_sender):
    activity_sender.include_mention = True
    activity_sender.init_create_note(lambda x: {**x})
    assert activity_sender.note.get("to") == [
        "https://www.w3.org/ns/activitystreams#Public"
    ]

    activity_sender.bovine_actor.post = AsyncMock()
    activity_sender.bovine_actor.get = AsyncMock(
        return_value={"inbox": "https://remote.example/inbox"}
    )

    await activity_sender.send("https://remote.example")
    data = activity_sender.bovine_actor.post.await_args[0][1]

    assert set(data["to"]) == {
        "https://remote.example",
        "https://www.w3.org/ns/activitystreams#Public",
    }

    obj = data.get("object")
    assert len(obj["tag"]) == 1
    assert obj["tag"][0] == {"type": "Mention", "href": "https://remote.example"}


async def test_activity_sender_replace_https(activity_sender):
    activity_sender.replace_https_with_http = True
    activity_sender.init_create_note(lambda x: {**x})

    activity_sender.bovine_actor.post = AsyncMock()
    activity_sender.bovine_actor.get = AsyncMock(
        return_value={"inbox": "https://remote.example/inbox"}
    )

    await activity_sender.send("https://remote.example")

    activity_sender.bovine_actor.get.assert_awaited_once()
    assert activity_sender.bovine_actor.get.await_args == [("http://remote.example",)]
    activity_sender.bovine_actor.post.assert_awaited_once()

    recipient = activity_sender.bovine_actor.post.await_args[0][0]

    assert recipient == "http://remote.example/inbox"



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import aiohttp
import json
from dataclasses import dataclass
import bovine
import logging


from fediverse_pasture.types import ApplicationAdapterForLastActivity

logger = logging.getLogger(__name__)


@dataclass
class MisskeyApplication:
    """Used to query a misskey application"""

    domain: str
    username: str
    session: aiohttp.ClientSession | None = None

    misskey_id: str | None = None

    async def determine_actor_uri(self):
        actor_uri, _ = await bovine.clients.lookup_uri_with_webfinger(
            self.session, f"acct:{self.username}@{self.domain}", f"http://{self.domain}"
        )
        return actor_uri

    async def determine_misskey_id(self):
        response = await self.session.post(
            f"http://{self.domain}/api/users/search",
            json={"query": "actor"},
            headers={"content-type": "application/json"},
        )
        users = await response.json()

        for user in users:
            if user.get("host") == "pasture-one-actor":
                return user.get("id")

    async def public_posts(self):
        if not self.misskey_id:
            self.misskey_id = await self.determine_misskey_id()

        response = await self.session.post(
            "http://misskey/api/users/notes",
            json={"userId": self.misskey_id},
            headers={"Authorization": "Bearer token"},
        )
        return await response.json()

    async def user_post_with_object_id(self, object_id: str):
        notes = await self.public_posts()

        for data in notes:
            if object_id in json.dumps(data):
                return data

    async def last_activity(self) -> ApplicationAdapterForLastActivity:
        actor_uri = await self.determine_actor_uri()

        return ApplicationAdapterForLastActivity(
            actor_uri=actor_uri,
            fetch_activity=self.user_post_with_object_id,
            application_name="misskey",
        )



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

import pytest

from .parser import restrictions_for_actors_from_args


def test_restrictions_for_actors_from_args_base():
    restriction, only = restrictions_for_actors_from_args([], {})

    assert restriction is False


@pytest.mark.parametrize(
    "args",
    [{"only_alice": True}, {"only_bob": True}, {"only_alice": True, "only_bob": True}],
)
def test_restrictions_for_actors_from_args(args):
    restriction, only = restrictions_for_actors_from_args(["alice", "bob"], args)

    assert restriction
    assert set(only.keys()) == {"alice", "bob"}



import asyncio
import click

from . import handle_send_to
from .modifier import ModifierBuilder


@click.command()
@click.option(
    "--domain",
    default="http://pasture-one-actor",
    help="Domain the actor is served one",
)
@click.option("--text", help="Content of the message to be send")
@click.option("--input_name", help="Name of the fediverse-pasture-input to use")
@click.option("--input_number", type=int, help="Id of the input to use")
@click.option(
    "--mention", is_flag=True, default=False, help="triggers mentioning the user"
)
@click.option(
    "--replace_https", is_flag=True, default=False, help="replace https:// with http://"
)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="if true prints the resulting activity",
)
@click.argument("uri")
def send_to(
    domain,
    text,
    uri,
    input_name: str,
    input_number: int,
    mention: bool,
    replace_https: bool,
    verbose: bool,
):
    modifier = ModifierBuilder(
        text=text, input_name=input_name, input_number=input_number
    ).build()
    if not asyncio.run(handle_send_to(modifier, domain, uri, mention, verbose)):
        exit(1)

    if verbose:
        print("Activity send successfully`")


if __name__ == "__main__":
    send_to()



import asyncio
import json

from quart import Quart, make_response
from . import BlueprintForActors

import logging

logger = logging.getLogger(__name__)


def app_for_data_provider(data_provider):
    app = Quart(__name__)
    queue = asyncio.Queue()

    async def generator():
        yield """<html><body><h1>Inbox content</h1>\n"""
        while True:
            data = await queue.get()
            if data:
                yield f"""<p><pre>{json.dumps(data, indent=2)}</pre></p>\n"""

    async def on_inbox(data):
        logger.error(data)
        await queue.put(data)

    app.register_blueprint(
        BlueprintForActors(
            actors=[data_provider.one_actor], on_inbox=on_inbox
        ).blueprint
    )

    @app.get("/")
    async def inbox_display():
        response = await make_response(generator())
        response.timeout = None
        return response

    return app



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import logging
import aiohttp

from dataclasses import dataclass, field
from typing import List, Callable, Awaitable

from quart import Blueprint, request, send_from_directory
from bovine.utils import webfinger_response_json
from bovine.crypto.types import CryptographicIdentifier
from bovine.crypto import build_validate_http_signature

from fediverse_pasture.data_provider.models import ActorData
from fediverse_pasture.data_provider import bovine_actor_for_actor_data
from .utils import actor_object_to_public_key


logger = logging.getLogger(__name__)


def assets_blueprint_for_directory(directory: str) -> Blueprint:
    """Returns a [Blueprint][quart.blueprints.Blueprint] that serves the directory
    as static files. Files ending in `.jsonap` will be served with
    content-type `application/activity+json`.

    :param directory: Directory to serve static files from
    """

    blueprint = Blueprint("assets", __name__)

    @blueprint.get("/assets/<filename>")
    async def assets(filename):
        if filename.endswith(".jsonap"):
            return await send_from_directory(
                directory, filename, mimetype="application/activity+json"
            )
        return await send_from_directory(directory, filename)

    return blueprint


def scheme_for_request(request) -> str:
    scheme = request.scheme
    if request.headers.get("X-Forwarded-Proto") == "https":
        scheme = "https"
    return scheme


@dataclass
class BlueprintForActors:
    """Creates a blueprint to expose actors. This blueprint handles
    exposing the Actor objects and their Inboxes

    If `application_actor` is not set, it is set to the actor
    with name `actor` from the list `actors`."""

    actors: List[ActorData] = field(
        metadata={"description": """The actors to expose through the blueprint"""},
    )
    application_actor: ActorData | None = field(
        default=None,
        metadata={
            "description": """Can be used to specify the application
        actor. Otherwise, the actor from `actors` with name 'actor' is used."""
        },
    )
    on_inbox: Callable[[dict], Awaitable] | None = field(
        default=None,
        metadata={
            "description": """Awaited when a message is received in the inbox. The
        json content of the message is passed as a parameter."""
        },
    )

    def __post_init__(self):
        if self.application_actor is None:
            self.application_actor = self.actor_for_name("actor")

    def actor_for_name(self, name: str) -> ActorData | None:
        for actor in self.actors:
            if actor.actor_name == name:
                return actor
        return None

    def actor_for_user(self, user: str) -> ActorData | None:
        for actor in self.actors:
            if actor.user_part == user:
                return actor
        return None

    async def validate_request(self, request):
        logger.info("Validating request")
        application_actor_id = f"{scheme_for_request(request)}://{request.host}/actor"
        application, _ = bovine_actor_for_actor_data(
            application_actor_id, self.application_actor
        )
        async with aiohttp.ClientSession() as session:
            await application.init(session=session)

            async def key_retriever(key_id):
                logger.debug("Retrieving key for %s", key_id)
                result = await application.get(key_id)
                public_key, owner = actor_object_to_public_key(result, key_id)
                logger.debug("Got key %s for owner %s", public_key, owner)
                return CryptographicIdentifier.from_pem(public_key, owner)

            result = await build_validate_http_signature(key_retriever)(request)

            logger.debug("Got validation result %s", str(result))
            logger.debug(request.headers)
            logger.debug(request.full_path)
            return result

    @property
    def blueprint(self) -> Blueprint:
        """[Quart Blueprint][quart.blueprints.Blueprint] providing the endpoints

        ```plaintext
            - GET /.well-known/webfinger
            - GET /<actor_name
            - POST /<actor_name>/inbox
        ```
        """
        actor_blueprint = Blueprint("actors", __name__)

        @actor_blueprint.get("/.well-known/webfinger")
        async def webfinger():
            resource = request.args.get("resource")
            if not resource or not resource.startswith("acct:") or "@" not in resource:
                return "", 404

            acct = resource.removeprefix("acct:")
            user, domain = acct.split("@", 1)

            if domain != request.host:
                return "", 404

            actor = self.actor_for_user(user)
            if not actor:
                return "", 404
            actor_id = (
                f"{scheme_for_request(request)}://{request.host}/{actor.actor_name}"
            )
            return webfinger_response_json(resource, actor_id)

        @actor_blueprint.get("/<actor_name>")
        async def actor_get(actor_name):
            actor = self.actor_for_name(actor_name)
            if not actor:
                return "not found", 404
            if actor.requires_signed_get_for_actor:
                if not await self.validate_request(request):
                    return "unauthorized", 401
            actor_id = (
                f"{scheme_for_request(request)}://{request.host}/{actor.actor_name}"
            )
            _, data = bovine_actor_for_actor_data(actor_id, actor)
            return data.build(), 200, {"content-type": "application/activity+json"}

        @actor_blueprint.post("/<actor_name>/inbox")
        async def inbox_post(actor_name):
            actor = self.actor_for_name(actor_name)
            if not actor:
                return "not found", 404
            if actor.requires_signed_post_for_inbox:
                if not await self.validate_request(request):
                    return "unauthorized", 401

            if self.on_inbox:
                try:
                    data = await request.get_json()
                    await self.on_inbox(data)
                except Exception as e:
                    logger.warning("Something went wrong in inbox parsing")
                    logger.warning(repr(e))

            return "", 202

        @actor_blueprint.get("/<actor_name>/outbox")
        async def outbox(actor_name):
            return {
                "@context": "https://www.w3.org/ns/activitystreams",
                "id": f"http://pasture-one-actor/{actor_name}/outbox",
                "type": "OrderedCollection",
                "totalItems": 0,
            }, {"content-type": "application/activity"}

        return actor_blueprint



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

from typing import Tuple


def restrictions_for_actors_from_args(actor_names, args) -> Tuple[bool, dict]:
    only = {name: args.get(f"only_{name}") for name in actor_names}

    if any(only.get(name) for name in actor_names):
        return True, only

    return False, {}



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from quart import Quart

import pytest
from .http_signature import http_signature_blueprint


@pytest.fixture
async def test_client():
    app = Quart(__name__)
    app.register_blueprint(http_signature_blueprint)
    yield app.test_client()


async def test_get_no_signature(test_client):
    response = await test_client.get("/")

    assert response.status_code == 401

    result = await response.get_json()

    assert result["x error"] == "Signature header is missing"


async def test_post_with_signature(test_client):
    response = await test_client.post(
        "/",
        data="{}",
        headers={
            "signature": 'keyId="about:inline",headers=" host date",signature="1xx=="'
        },
    )

    assert response.status_code == 401



from urllib.parse import urlparse

from quart import Blueprint, request

from bovine.models import JrdData, JrdLink
from bovine.utils import pydantic_to_json


nodeinfo_blueprint = Blueprint("nodeinfo", __name__)


@nodeinfo_blueprint.get("/.well-known/nodeinfo")
async def nodeinfo() -> tuple[dict, int]:
    """Returns the JRD corresponding to `/.well-known/nodeinfo`"""
    path = urlparse(request.url)
    nodeinfo = JrdLink(
        rel="http://nodeinfo.diaspora.software/ns/schema/2.0",
        href=f"{path.scheme}://{path.netloc}/.well-known/nodeinfo2_0",
    )

    return (
        pydantic_to_json(JrdData(links=[nodeinfo])),
        200,
        {"content-type": "application/jrd+json"},
    )


@nodeinfo_blueprint.get("/.well-known/nodeinfo2_0")
async def nodeinfo_response() -> dict:
    user_count = 0
    user_stat = {
        "total": user_count,
        "activeMonth": user_count,
        "activeHalfyear": user_count,
    }

    return {
        "metadata": {},
        "openRegistrations": False,
        "protocols": ["activitypub"],
        "services": {"inbound": [], "outbound": []},
        "software": {"name": "fediverse-pasture", "version": "0"},
        "usage": {"users": user_stat},
        "version": "2.0",
    }



import pytest

from quart import Quart

from .nodeinfo_blueprint import nodeinfo_blueprint


@pytest.fixture
def app_client():
    app = Quart(__name__)
    app.register_blueprint(nodeinfo_blueprint)

    yield app.test_client()


async def test_url(app_client):
    response = await app_client.get("/.well-known/nodeinfo")

    assert response.status_code == 200

    data = await response.json

    assert data["links"][0]["href"] == "http://localhost/.well-known/nodeinfo2_0"
